--------------------------------------------------------------------------------
-- Generic deriving with support for 1 parameter (GHC.Generics)
--------------------------------------------------------------------------------

{-# LANGUAGE DataKinds              #-}
{-# LANGUAGE PolyKinds              #-}
{-# LANGUAGE TypeFamilies           #-}
{-# LANGUAGE GADTs                  #-}
{-# LANGUAGE FlexibleContexts       #-}
{-# LANGUAGE FlexibleInstances      #-}
{-# LANGUAGE ScopedTypeVariables    #-}
{-# LANGUAGE DefaultSignatures      #-}
{-# LANGUAGE TypeOperators          #-}

module Library.IG where

--------------------------------------------------------------------------------
-- Un
--------------------------------------------------------------------------------

data Un s = 
  -- Void
    VV

  -- Unit
  | UU

  -- Constants (either other parameters or recursion into types of kind *)
  | KK s
  
  -- Sum, product
  | Un s :+:  Un s
  | Un s :**: Un s

--------------------------------------------------------------------------------
-- Interpretation (as a GADT)
--------------------------------------------------------------------------------

data In (u :: Un *) :: * where

  -- No interpretation for VV, as it shouldn't map to any value
  
  -- Unit
  U     :: In UU
  
  -- Constants
  K     :: x -> In (KK x)
  
  -- Sum
  L     :: In f -> In (f :+: g)
  R     :: In g -> In (f :+: g)

  -- Product
  (:*:) :: In f -> In g -> In (f :**: g)

--------------------------------------------------------------------------------
-- Conversions to/from user datatypes
--------------------------------------------------------------------------------

class Representable (a :: *) where
  type Rep a :: Un *

  from :: a -> In (Rep a)
  to   :: In (Rep a) -> a

--------------------------------------------------------------------------------
-- Example function: equality
--------------------------------------------------------------------------------

class GEq' (a :: Un *) where
  geq' :: In a -> In a -> Bool

instance GEq' UU where
  geq' _ _ = True

instance (GEq c) => GEq' (KK c) where
  geq' (K a) (K b) = geq a b

instance (GEq' a, GEq' b) => GEq' (a :+: b) where
  geq' (L a) (L b) = geq' a b
  geq' (R a) (R b) = geq' a b
  geq' _      _      = False

instance (GEq' a, GEq' b) => GEq' (a :**: b) where
  geq' (a1 :*: b1) (a2 :*: b2) = geq' a1 a2 && geq' b1 b2


-- User-facing class
class GEq (a :: *) where 
  geq :: a -> a -> Bool

  default geq :: (Representable a, GEq' (Rep a)) => a -> a -> Bool
  geq x y = geq' (from x) (from y)


-- Base types instances
instance GEq Char   where geq = (==)
instance GEq Int    where geq = (==)
instance GEq Float  where geq = (==)


-- The examples below are all commented because they introduce IG
-- representations, and that overlaps with the most general IG representation
-- of D2IG. Basically, what is below is the code that no longer has to be
-- written (or generated with TH).

--------------------------------------------------------------------------------
-- Example encoding: lists
--------------------------------------------------------------------------------

-- instance Representable [a] where
  -- type Rep [a] = UU :+: (KK a :**: KK [a])

  -- from []    = L U
  -- from (h:t) = R (K h :*: K t)

  -- to (L U)             = []
  -- to (R (K h :*: K t)) = h:t

-- instance (GEq a) => GEq [a]

--------------------------------------------------------------------------------
-- Example encoding: rose trees
--------------------------------------------------------------------------------

data RTree a = RTree a [RTree a] deriving Show

-- instance Representable (RTree a) where
  -- type Rep (RTree a) = KK a :**: KK [RTree a]

  -- from (RTree x xs)   = K x :*: K xs
  -- to   (K x :*: K xs) = RTree x xs

-- instance (GEq a) => GEq (RTree a)

--------------------------------------------------------------------------------
-- Tests
--------------------------------------------------------------------------------

-- tgeq1 = geq [1] [2::Int]
-- tgeq2 = geq [1] [1::Int]
-- tgeq3 = geq (RTree (0::Int) [RTree 1 []]) (RTree 0 [RTree 1 [], RTree 2 []])
